# -*- coding: utf-8 -*-

import os
from enum import IntEnum
from enum import IntFlag
from enum import unique
from functools import cached_property
from functools import partial
from typing import Callable
from typing import NamedTuple # there should probably be a
                              # lot more usage of this lol
from typing import Optional

from cmyui import utils
from cmyui import logging

__all__ = ('TimingPoint', 'SampleSet', 'HitSample',
           'HitObject', 'HitCircle', 'Slider', 'Spinner', 'ManiaHold',
           'Beatmap', 'CurveType', 'HitSound', 'Colour', 'OverlayPosition',
           'Event', 'Background', 'Video', 'Break')

"""\
a (rather slow) complete osu! beatmap parser, with basic safety.

this is somewhat older code that's been revamped a few times,
some of this should probably be written into c to speed things up.
it has complete coverage of modern .osu beatmap files, with an
api under active development alongside https://github.com/cmyui/gulag.

Basic usage:
```
  b = Beatmap.from_file('1234567.osu')
  if not b: # file not found
    ...

  # now you have basically everything
  # you can think of from a beatmap.

  # most complex structures are classes,
  # with many attributes within them..
  for tp in b.timing_points:
    print(tp.time, tp.beat_length, tp.meter, ...)
    ...

  for obj in b.hit_objects:
    print(obj.x, obj.y, obj.time, obj.hit_sample.volume, ...)
    ...

  # and much more.. for now the documentation will stay
  # developer-friendly and perhaps get better in the future,
  # for the time being, read the code ;)
  # it's getting simpler i promise
  ...
```

soon i'll be adding some kind of simulator that can play
through the map and achieve a perfect score.. it would be
useful for scorev2 detection on gulag too.. could finally
be rid of the scorev2 bugginess :o
"""

# TODO: some of the classmethod usage in this is pretty nasty,
# seems like i didn't fully understand it at the time of writing.

class TimingPoint:
    def __init__(
        self, time: int, beat_length: float,
        meter: int, sample_set: int, sample_index: int,
        volume: int, uninherited: bool, effects: int
    ) -> None:
        self.time = time
        self.beat_length = beat_length
        self.meter = meter
        self.sample_set = sample_set
        self.sample_index = sample_index
        self.volume = volume
        self.uninherited = uninherited
        self.effects = effects

    @cached_property
    def bpm(self) -> int:
        return 1 / self.beat_length * 1000 * 60

    @classmethod
    def from_str(cls, s: str) -> 'TimingPoint':
        if len(tp_split := s.split(',')) != 8:
            return

        isfloat_n = partial(
            utils._isdecimal,
            _float = True,
            _negative = True
        )

        # make sure all params are at least floats
        if all([isfloat_n(x) for x in tp_split]):
            return cls(
                time=int(tp_split[0]),
                beat_length=float(tp_split[1]),
                meter=int(tp_split[2]),
                sample_set=int(tp_split[3]),
                sample_index=int(tp_split[4]),
                volume=int(tp_split[5]),
                uninherited=tp_split[6] == '1',
                effects=int(tp_split[7])
            )

@unique
class SampleSet(IntEnum):
    NONE = 0
    NORMAL = 1
    SOFT = 2
    DRUM = 3

    def __str__(self) -> str:
        return self.name.lower()

class HitSample:
    def __init__(
        self, normal_set: SampleSet, addition_set: SampleSet,
        index: int, volume: int, filename: str
    ) -> None:
        self.normal_set = normal_set
        self.addition_set = addition_set
        self.index = index
        self.volume = volume
        self.filename = filename

    @classmethod
    def from_str(cls, s: str) -> 'HitSample':
        if len(hs_split := s.split(':')) != 5:
            return

        if all(map(str.isdecimal, hs_split[:-1])):
            return cls(
                normal_set=SampleSet(int(hs_split[0])),
                addition_set=SampleSet(int(hs_split[1])),
                index=int(hs_split[2]),
                volume=int(hs_split[3]),
                filename=hs_split[4]
            )

""" # slow :(
@unique
class ObjectType(IntFlag):
    HIT_CIRCLE = 1 << 0
    SLIDER = 1 << 1
    NEW_COMBO = 1 << 2
    SPINNER = 1 << 3

    # a bit weird?
    # maybe it should be a property
    # where it just checks all the bits
    # or soemthing.. will think abt this
    SKIP_ONE = 1 << 4
    SKIP_TWO = 1 << 5
    SKIP_THREE = 1 << 6

    MANIA_HOLD = 1 << 7
"""

HIT_CIRCLE = 1 << 0
SLIDER = 1 << 1
NEW_COMBO = 1 << 2
SPINNER = 1 << 3

SKIP_ONE = 1 << 4
SKIP_TWO = 1 << 5
SKIP_THREE = 1 << 6

MANIA_HOLD = 1 << 7

@unique
class HitSound(IntFlag):
    NORMAL = 1 << 0
    WHISTLE = 1 << 1
    FINISH = 1 << 2
    CLAP = 1 << 3

    def __str__(self) -> str:
        return self.name.lower()

class HitObject:
    def __init__(
        self,
        x: int, y: int, time: int,
        hit_sound: HitSound,
        hit_sample: Optional[HitSample] = None
    ) -> None:
        self.x = x
        self.y = y
        self.time = time

        # very closely related
        # XXX: i could actually refactor these to
        # be in a single class (and may), but i think
        # it might confuse people who are reading osu!'s
        # implementation to make sense of it? idk man lol
        self.hit_sound = hit_sound
        self.hit_sample = hit_sample

    @staticmethod
    def from_str(s: str) -> 'HitObject':
        if len(args := s.split(',', 5)) != 6:
            return

        # make sure all params so far are integral
        if all(map(str.isdecimal, args[:-1])):
            # parse common items from hit object
            # the first 5 params of any hitobject
            type_ = int(args[3])

            # hit sample not read yet, it may be at the
            # end of the args list depending on the type.

            if type_ & HIT_CIRCLE:
                cls = HitCircle
            elif type_ & SLIDER:
                cls = Slider
            elif type_ & SPINNER:
                cls = Spinner
            elif type_ & MANIA_HOLD:
                cls = ManiaHold
            else:
                logging.log(
                    f'Unknown hit object type {type_}',
                    logging.Ansi.LYELLOW
                )
                return

            return cls.from_str(
                s=args[5],
                x=int(args[0]),
                y=int(args[1]),
                time=int(args[2]),
                hit_sound=HitSound(int(args[4]))
            ) # pass rest of the args

class HitCircle(HitObject):
    # hitcircle is simple, nothing extra,
    # so we don't have to write constructor

    @classmethod
    def from_str(cls, s: str, **kwargs):
        if s != '0:0:0:0:':
            kwargs['hit_sample'] = HitSample.from_str(s)

        return cls(**kwargs)

@unique
class CurveType(IntEnum):
    Bezier = 0
    Catmull = 1
    Linear = 2
    Perfect = 3

    #def __str__(self) -> str:
    #    # first char of character
    #    return self.name[0]

class Slider(HitObject):
    def __init__(
        self, curve_type: CurveType,
        curve_points: list[str],
        slides: int,
        length: float,
        edge_sounds: list[int] = [],
        edge_sets: list[list[int, int]] = [],
        **kwargs
    ) -> None:
        self.curve_type = curve_type
        self.curve_points = curve_points
        self.slides = slides
        self.length = length
        self.edge_sounds = edge_sounds
        self.edge_sets = edge_sets

        super().__init__(**kwargs)

    @classmethod
    def from_str(cls, s: str, **kwargs):
        if len(split := s.split(',')) < 3:
            return

        _curve, _slides, _slen, *_extra = split
        ctype, cpoints = _curve.split('|', 1)

        if _extra:
            assert len(_extra) == 3
            kwargs |= {
                'edge_sounds': list(map(int, _extra[0].split('|'))),
                'edge_sets': [x.split(':', 1) for x in _extra[1].split('|')],
                'hit_sample': HitSample.from_str(_extra[2])
            }

        return cls(
            curve_type={
                'B': CurveType.Bezier, 'C': CurveType.Catmull,
                'L': CurveType.Linear, 'P': CurveType.Perfect
            }[ctype],
            curve_points=cpoints.split('|'),
            slides=int(_slides),
            length=float(_slen),
            **kwargs
        )

class Spinner(HitObject):
    def __init__(self, end_time: int, **kwargs) -> None:
        self.end_time = end_time

        super().__init__(**kwargs)

    @classmethod
    def from_str(cls, s, **kwargs):
        if len(split := s.split(',')) != 2:
            return

        if split[0].isdecimal():
            if split[1] != '0:0:0:0:':
                kwargs |= {'hit_sample': HitSample.from_str(split[1])}

            return cls(end_time=int(split[0]), **kwargs)

class ManiaHold(HitObject):
    def __init__(self, end_time: int, **kwargs) -> None:
        self.end_time = end_time

        super().__init__(**kwargs)

        # `self.x` determines the column the hold will be in;
        # it can be determined with floor(x * columnCount / 512)
        # clamped between 0 and columnCount - 1
        # y will default to the centre of playfield, 192

    @classmethod
    def from_str(cls, s, **kwargs):
        if len(split := s.split(':')) != 2:
            return

        if split[0].isdecimal():
            if split[1] != '0:0:0:0:':
                kwargs |= {'hit_sample': HitSample.from_str(split[1])}

            return cls(end_time=int(split[0]), **kwargs)

class Colour(NamedTuple):
    r: int
    g: int
    b: int

    @classmethod
    def from_str(cls, s: str):
        if len(split := s.split(',')) != 3:
            return

        if all(map(str.isdecimal, split)):
            return cls(
                r=int(split[0]),
                g=int(split[1]),
                b=int(split[2])
            )

@unique
class OverlayPosition(IntEnum):
    No_Change = 0
    Below = 1
    Above = 2

class Event:
    def __init__(self, start_time: int) -> None:
        self.start_time = start_time

    @staticmethod
    def from_str(s: str):
        if len(split := s.split(',', 2)) != 3:
            return

        if split[1].isdecimal():
            # a bit unsafe but w/e
            _type = split[0]

            if _type.isdecimal():
                ev_map = (Background, Video, Break)
                _type = int(_type)
            else:
                ev_map = {'Video': Video, 'Break': Break}

            cls = ev_map[_type]

            return cls.from_str(split[2], start_time=int(split[1]))

class Background(Event):
    def __init__(
        self, filename: str,
        x_offset: Optional[int] = None,
        y_offset: Optional[int] = None,
        **kwargs
    ) -> None:
        self.filename = filename
        self.x_offset = x_offset
        self.y_offset = y_offset

        super().__init__(**kwargs)

    @classmethod
    def from_str(cls, s: str, **kwargs):
        split = s.split(',', 2)
        lsplit = len(split)

        if lsplit == 3:
            x_off, y_off = split[1:3]
            if not (x_off.isdecimal() and y_off.isdecimal()):
                return

            kwargs |= {'x_offset': int(x_off),
                       'y_offset': int(y_off)}
        elif lsplit != 1:
            raise Exception('Invalid arg count for a background.')

        kwargs |= {
            'filename': split[0].strip('"')
        }

        return cls(**kwargs)

# they're literally identical, since i'm
# still parsing the time for backgrounds lol
Video = type('Video', Background.__bases__, dict(Background.__dict__))

class Break(Event):
    def __init__(self, end_time: int, **kwargs) -> None:
        self.end_time = end_time

        super().__init__(**kwargs)

    @classmethod
    def from_str(cls, s: str, **kwargs):
        if s.isdecimal():
            return cls(end_time=int(s), **kwargs)

# TODO: storyboards?

class Beatmap:
    def __init__(self) -> None:
        self.file_version: Optional[int] = None

        """ general """
        self.audio_filename: Optional[str] = None
        self.audio_leadin: Optional[int] = None
        #self.audio_hash: Optional[str] = None # deprecated
        self.preview_time: Optional[int] = None
        self.countdown: Optional[int] = None
        self.sample_set: Optional[str] = None
        self.stack_leniency: Optional[float] = None
        self.mode: Optional[int] = None
        self.letterbox_in_breaks: Optional[bool] = None
        #self.storyfire_in_front: Optional[bool] = None # deprecated
        self.use_skin_sprites: Optional[bool] = None
        #self.always_show_playback: Optional[bool] = None # deprecated
        self.overlay_position: Optional[str] = None
        self.skin_preference: Optional[str] = None
        self.epilepsy_warning: Optional[bool] = None
        self.countdown_offset: Optional[int] = None
        self.special_style: Optional[bool] = None
        self.widescreen_storyboard: Optional[bool] = None
        self.samples_match_playback_rate: Optional[bool] = None

        """ editor """
        self.bookmarks: Optional[list[int]] = None
        self.distance_spacing: Optional[float] = None
        self.beat_divisor: Optional[float] = None
        self.grid_size: Optional[int] = None
        self.timeline_zoom: Optional[float] = None

        """ metadata """
        self.title: Optional[str] = None
        self.title_unicode: Optional[str] = None
        self.artist: Optional[str] = None
        self.artist_unicode: Optional[str] = None
        self.creator: Optional[str] = None
        self.version: Optional[str] = None
        self.source: Optional[str] = None
        self.tags: Optional[list[str]] = None
        self.id: Optional[int] = None
        self.set_id: Optional[int] = None

        """ difficulty """
        self.diff_hp: Optional[float] = None
        self.diff_cs: Optional[float] = None
        self.diff_od: Optional[float] = None
        self.diff_ar: Optional[float] = None
        self.slider_multiplier: Optional[float] = None
        self.slider_tick_rate: Optional[float] = None

        """ events """
        self.backgrounds: Optional[list[Background]] = None
        self.breaks: Optional[list[Break]] = None
        self.videos: Optional[list[Video]] = None
        #self.storyboards: Optional[list[Storyboard]] = None # TODO

        """ timing points, colours & hit objects """
        self.timing_points: Optional[list[TimingPoint]] = None
        self.colours: Optional[dict[str, Colour]] = None
        self.hit_objects: Optional[list[HitObject]] = None

        """ internal reader use only """
        self._data: Optional[str] = None
        self._offset: Optional[int] = None

    def __repr__(self) -> str:
        return (
            '{artist} - {title} ({creator}) [{version}]'
        ).format(**self.__dict__)

    @property
    def data(self) -> str:
        # return all data starting from
        # the internal offset of the reader.
        return self._data[self._offset:]

    @classmethod
    def from_file(cls, filename: str) -> 'Beatmap':
        if not os.path.exists(filename):
            return

        b = cls()

        with open(filename, 'r') as f:
            l: list[str] = []

            # remove any commented-out lines
            for line in f.readlines():
                if not line.startswith('//'):
                    l.append(line)

            b._data = ''.join(l)
            b._offset = 0

        b._parse()
        return b

    def _parse(self) -> None:
        sec_start = self.data.find('\n\n')
        self._offset += len('osu file format v')

        ver_str = self.data[:sec_start - self._offset]
        if not ver_str.isdecimal():
            raise Exception('Invalid file format.')

        self.file_version = int(ver_str)

        for name, func in (
            ('General', self._parse_general),
            ('Editor', self._parse_editor),
            ('Metadata', self._parse_metadata),
            ('Difficulty', self._parse_difficulty),
            ('Events', self._parse_events),
            ('TimingPoints', self._parse_timing_points),
            ('Colours', self._parse_colours),
            ('HitObjects', self._parse_hit_objects)
        ):
            self._parse_section(name, func)

        # TODO
        # parsing file completed, now
        # construct any additional info
        # now that we have everything.
        #self._parse_end()

    def _parse_section(
        self, name: str,
        parse_method: Callable[['Beatmap'], None]
    ) -> None:
        to_find = f'\n\n[{name}]\n'
        offs = self.data.find(to_find)

        if offs == -1:
            # skip any sections not found - the beatmap
            # object will simply have `None` attributes
            # if not parsed from the file.
            return

        self._offset += offs + len(to_find)
        parse_method()

    def _parse_general(self) -> None:
        end_of_section = self.data.find('\n\n')

        for line in self.data[:end_of_section].splitlines():
            key, val = line.split(': ', maxsplit=1)

            # how should i clean this.. lol

            if key == 'AudioFilename':
                self.audio_filename = val
            elif key == 'AudioLeadIn':
                if val.isdecimal():
                    self.audio_leadin = int(val)
            # deprecated
            #elif key == 'AudioHash':
            #    self.audio_hash = val
            elif key == 'PreviewTime':
                if val.isdecimal():
                    self.preview_time = int(val)
            elif key == 'Countdown':
                if val.isdecimal():
                    self.countdown = int(val)
            elif key == 'SampleSet':
                if val in ('Normal', 'Soft', 'Drum'):
                    self.sample_set = {
                        'Normal': SampleSet.NORMAL,
                        'Soft': SampleSet.SOFT,
                        'Drum': SampleSet.DRUM
                    }[val]
            elif key == 'StackLeniency':
                if utils._isdecimal(val, _float=True):
                    self.stack_leniency = float(val)
            elif key == 'Mode':
                if val.isdecimal():
                    self.mode = int(val)
            elif key == 'LetterboxInBreaks':
                self.letterbox_in_breaks = val == '1'
            # deprecated
            #elif key == 'StoryFireInFront':
            #    self.story_fire_in_front = val == '1'
            elif key == 'UseSkinSprites':
                self.use_skin_sprites = val == '1'
            # deprecated
            #elif key == 'AlwaysShowPlayfield':
            #    self.always_show_playfield = val == '1'
            elif key == 'OverlayPosition':
                if val in ('NoChange', 'Below', 'Above'):
                    self.overlay_position = {
                        'NoChange': OverlayPosition.No_Change,
                        'Below': OverlayPosition.Below,
                        'Above': OverlayPosition.Above
                    }[val]
            elif key == 'SkinPreference':
                self.skin_preference = val
            elif key == 'EpilepsyWarning':
                self.epilepsy_warning = val == '1'
            elif key == 'CountdownOffset':
                if val.isdecimal():
                    self.countdown_offset = int(val)
            elif key == 'SpecialStyle':
                self.special_style = val == '1'
            elif key == 'WidescreenStoryboard':
                self.widescreen_storyboard = val == '1'
            elif key == 'SamplesMatchPlaybackRate':
                self.samples_match_playback_rate = val == '1'
            else:
                logging.log(
                    f'Unknown [General] key {key}',
                    logging.Ansi.LYELLOW
                )

        self._offset += end_of_section

    def _parse_editor(self) -> None:
        e_end = self.data.find('\n\n')

        for line in self.data[:e_end].splitlines():
            key, val = line.split(': ', maxsplit=1)

            if key == 'Bookmarks':
                bookmarks_str = val.split(',')
                if all(map(str.isdecimal, bookmarks_str)):
                    self.bookmarks = list(map(int, bookmarks_str))
            elif key == 'DistanceSpacing':
                if utils._isdecimal(val, _float=True):
                    self.distance_spacing = float(val)
            elif key == 'BeatDivisor':
                if utils._isdecimal(val, _float=True):
                    self.beat_divisor = float(val)
            elif key == 'GridSize':
                if val.isdecimal():
                    self.grid_size = int(val)
            elif key == 'TimelineZoom':
                if utils._isdecimal(val, _float=True):
                    self.timeline_zoom = float(val)
            else:
                logging.log(
                    f'Unknown [Editor] key {key}',
                    logging.Ansi.LYELLOW
                )

        self._offset += e_end

    def _parse_metadata(self) -> None:
        m_end = self.data.find('\n\n')

        for line in self.data[:m_end].splitlines():
            key, val = line.split(':', maxsplit=1)

            if key == 'Title':
                self.title = val
            elif key == 'TitleUnicode':
                self.title_unicode = val
            elif key == 'Artist':
                self.artist = val
            elif key == 'ArtistUnicode':
                self.artist_unicode = val
            elif key == 'Creator':
                self.creator = val
            elif key == 'Version':
                self.version = val
            elif key == 'Source':
                self.source = val
            elif key == 'Tags':
                self.tags = val.split(' ')
            elif key == 'BeatmapID':
                if val.isdecimal():
                    self.id = int(val)
            elif key == 'BeatmapSetID':
                if val.isdecimal():
                    self.set_id = int(val)
            else:
                logging.log(
                    f'Unknown [Metadata] key {key}',
                    logging.Ansi.LYELLOW
                )

        self._offset += m_end

    def _parse_difficulty(self) -> None:
        d_end = self.data.find('\n\n')

        for line in self.data[:d_end].splitlines():
            key, val = line.split(':', maxsplit=1)

            # all diff params should be float
            if utils._isdecimal(val, _float=True):
                if key == 'HPDrainRate':
                    self.diff_hp = float(val)
                elif key == 'CircleSize':
                    self.diff_cs = float(val)
                elif key == 'OverallDifficulty':
                    self.diff_od = float(val)
                elif key == 'ApproachRate':
                    self.diff_ar = float(val)
                elif key == 'SliderMultiplier':
                    self.slider_multiplier = float(val)
                elif key == 'SliderTickRate':
                    self.slider_tick_rate = float(val)
                else:
                    logging.log(
                        f'Unknown [Difficulty] key {key}',
                        logging.Ansi.LYELLOW
                    )

        self._offset += d_end

    def _parse_events(self) -> None:
        self.backgrounds = []
        self.videos = []
        self.breaks = []

        # this is actually events, backgrounds,
        # videos, breaks, and storyboards.. lol
        ev_end = self.data.find('\n\n')

        for line in self.data[:ev_end].splitlines():
            ev = Event.from_str(line)

            if isinstance(ev, Background):
                self.backgrounds.append(ev)
            elif isinstance(ev, Video):
                self.videos.append(ev)
            elif isinstance(ev, Break):
                self.breaks.append(ev)
            else:
                breakpoint()

        self._offset += ev_end

    def _parse_timing_points(self) -> None:
        self.timing_points = []

        # find the end of the timing points section
        tp_end = self.data.find('\n\n')

        # iterate through each line, parsing
        # the lines into timing point objects.
        for line in self.data[:tp_end].splitlines():
            if not (tp := TimingPoint.from_str(line)):
                logging.printc(
                    f'Failed to parse timing point? "{line}"',
                    logging.Ansi.RED
                )
                continue

            self.timing_points.append(tp)

        self._offset += tp_end

    def _parse_colours(self) -> None:
        self.colours = {}

        # find the end of the colours section
        cl_end = self.data.find('\n\n')

        for line in self.data[:cl_end].splitlines():
            key, val = line.split(' : ', maxsplit=1)

            if colour := Colour.from_str(val):
                # add to beatmap's colours
                self.colours[key] = colour

        self._offset += cl_end

    def _parse_hit_objects(self) -> None:
        self.hit_objects = []

        # iterate through each line, parsing
        # the lines into hit object objects
        for line in self.data.splitlines():
            if not (ho := HitObject.from_str(line)):
                logging.printc(
                    f'Failed to parse hit object? "{line}"',
                    logging.Ansi.RED
                )
                continue

            self.hit_objects.append(ho)

        self._offset += len(self.data)

if __name__ == '__main__':
    os.chdir(os.path.dirname(os.path.realpath(__file__)))

    import time
    st = time.time_ns()
    bmap = Beatmap.from_file('tstmts.osu')
    elapsed = utils.magnitude_fmt_time(time.time_ns()-st)
    print(f'Parsed {bmap} in {elapsed}.')

    #import cProfile
    #cProfile.run("Beatmap.from_file('tstmts.osu')", filename='bmap.stats')



# -*- coding: utf-8 -*-

import sys
from datetime import tzinfo
from enum import IntEnum
from functools import cache
from functools import lru_cache
from typing import Union
from typing import Optional
from typing import overload
from zoneinfo import ZoneInfo

from .utils import get_timestamp

__all__ = ('Ansi', 'AnsiRGB', 'printc', 'set_timezone', 'log')

class Ansi(IntEnum):
    # Default colours
    BLACK   = 30
    RED     = 31
    GREEN   = 32
    YELLOW  = 33
    BLUE    = 34
    MAGENTA = 35
    CYAN    = 36
    WHITE   = 37

    # Light colours
    GRAY     = 90
    LRED     = 91
    LGREEN   = 92
    LYELLOW  = 93
    LBLUE    = 94
    LMAGENTA = 95
    LCYAN    = 96
    LWHITE   = 97

    RESET = 0

    @cache
    def __repr__(self) -> str:
        return f'\x1b[{self.value}m'

class AnsiRGB:
    @overload
    def __init__(self, rgb: int) -> None: ...
    @overload
    def __init__(self, r: int, g: int, b: int) -> None: ...

    def __init__(self, *args) -> None:
        largs = len(args)

        if largs == 3:
            # r, g, b passed.
            self.r, self.g, self.b = args
        elif largs == 1:
            # passed as single argument
            rgb = args[0]
            self.b = rgb & 0xff
            self.g = (rgb >> 8) & 0xff
            self.r = (rgb >> 16) & 0xff
        else:
            raise ValueError('Incorrect params for AnsiRGB.')

    @lru_cache(maxsize=64)
    def __repr__(self) -> str:
        return f'\x1b[38;2;{self.r};{self.g};{self.b}m'

Ansi_T = Union[Ansi, AnsiRGB]

stdout_write = sys.stdout.write
stdout_flush = sys.stdout.flush

_gray = repr(Ansi.GRAY)
_reset = repr(Ansi.RESET)

def printc(s: str, col: Ansi_T, end: str = '\n') -> None:
    """Print a string, in a specified ansi colour."""
    stdout_write(f'{col!r}{s}{_reset}{end}')
    stdout_flush()

_log_tz = ZoneInfo('America/Toronto') # default
def set_timezone(tz: tzinfo) -> None:
    global _log_tz
    _log_tz = tz

def log(msg: str, col: Optional[Ansi_T] = None,
        fd: Optional[str] = None, end: str = '\n') -> None:
    """\
    Print a string, in a specified ansi colour with timestamp.

    Allows for the functionality to write to a file as
    well by passing the filepath with the `fd` parameter.
    """

    ts_short = get_timestamp(full=False, tz=_log_tz)

    if col:
        stdout_write(f'{_gray}[{ts_short}] {col!r}{msg}{_reset}{end}')
    else:
        stdout_write(f'{_gray}[{ts_short}]{_reset} {msg}{end}')

    stdout_flush()

    if fd:
        # log simple ascii output to fd.
        with open(fd, 'a+') as f:
            f.write(f'[{get_timestamp(full=True, tz=_log_tz)}] {msg}\n')



# -*- coding: utf-8 -*-

# Domain-based asynchronous server implementation, written from
# sockets, mostly with https://github.com/cmyui/gulag in mind.

import asyncio
import gzip
import http
import importlib
import inspect
import os
import re
import select
import signal
import socket
import sys
from functools import wraps
from time import perf_counter as clock
from time import perf_counter_ns as clock_ns
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Iterable
from typing import Optional
from typing import Union

from .logging import Ansi
from .logging import AnsiRGB
from .logging import log
from .logging import printc
from .utils import magnitude_fmt_time

__all__ = (
    'Address',
    'STATUS_LINES',
    'ratelimit',
    'Connection',
    'RouteMap',
    'Domain',
    'Server'
)

STATUS_LINES = {
    c.value: f'HTTP/1.1 {c.value} {c.phrase.upper()}'
    for c in http.HTTPStatus
}

def ratelimit(period: int, max_count: int,
              default_return: Optional[Any] = None
             ) -> Callable:
    """Utility decorator for global ratelimiting."""
    period = period
    max_count = max_count
    default_return = default_return

    last_reset = 0
    num_calls = 0

    def decorate(f: Callable) -> Callable:
        # TODO: not an extra 18 lines for 6 char change
        if inspect.iscoroutinefunction(f):
            async def wrapper(*args, **kwargs) -> Optional[Any]:
                nonlocal period, max_count, last_reset, num_calls

                elapsed = clock() - last_reset
                period_remaining = period - elapsed

                if period_remaining <= 0:
                    num_calls = 0
                    last_reset = clock()

                num_calls += 1

                if num_calls > max_count:
                    # call ratelimited.
                    return default_return

                return await f(*args, **kwargs)
        else:
            def wrapper(*args, **kwargs) -> Optional[Any]:
                nonlocal period, max_count, last_reset, num_calls

                elapsed = clock() - last_reset
                period_remaining = period - elapsed

                if period_remaining <= 0:
                    num_calls = 0
                    last_reset = clock()

                num_calls += 1

                if num_calls > max_count:
                    # call ratelimited.
                    return default_return

                return f(*args, **kwargs)

        return wraps(f)(wrapper)
    return decorate

# (host, port) for inet; sockpath for unix
Address = Union[tuple[str, int], str]
Hostname_Types = Union[str, Iterable[str], re.Pattern]

REQUEST_LINE_RGX = re.compile(
    r'^(?P<cmd>GET|HEAD|POST|PUT|DELETE|PATCH|OPTIONS) '
    r'(?P<path>/[^? ]*)(?P<args>\?[^ ]+)? ' # cursed?
    r'HTTP/(?P<httpver>1\.0|1\.1|2\.0|3\.0)$'
)

class CaseInsensitiveDict(dict):
    """A dictionary with case insensitive keys."""
    def __init__(self, *args, **kwargs) -> None:
        self._keystore = {}
        d = dict(*args, **kwargs)
        for k in d.keys():
            self._keystore[k.lower()] = k
        return super().__init__(*args, **kwargs)

    def __setitem__(self, k, v) -> None:
        self._keystore[k.lower()] = k
        return super().__setitem__(k, v)

    def __getitem__(self, k) -> str:
        k_lower = k.lower()
        if k_lower in self._keystore:
            k = self._keystore[k_lower]
        return super().__getitem__(k)

    def __contains__(self, k: str) -> bool:
        k_lower = k.lower()
        if k_lower in self._keystore:
            k = self._keystore[k_lower]
        return super().__contains__(k)

    def get(self, k, failobj=None) -> Optional[str]:
        k_lower = k.lower()
        if k_lower in self._keystore:
            k = self._keystore[k_lower]
        return super().get(k, failobj)

class Connection:
    __slots__ = (
        'client',

        # Request params
        'headers', 'body', 'cmd',
        'path', 'raw_path', 'httpver',

        'args', 'multipart_args', 'files',

        '_buf',

        # Response params
        'resp_code', 'resp_headers'
    )

    def __init__(self, client: socket.socket) -> None:
        self.client = client

        # Request params
        self.headers = CaseInsensitiveDict()
        self.body: Optional[memoryview] = None
        self.cmd = None
        self.path = None
        self.raw_path = None
        self.httpver = 0.0

        self.args = {}
        self.multipart_args = {}

        self.files = {}

        # Response params
        self.resp_code = 200
        self.resp_headers = {}

        self._buf = bytearray()

    """ Request methods """

    def _parse_urlencoded(self, data: bytes) -> None:
        for a_pair in data.split('&'):
            a_key, a_val = a_pair.split('=', 1)
            self.args[a_key] = a_val

    def _parse_headers(self, data: str) -> None:
        """Parse the http headers from the internal body."""
        # split up headers into http line & header lines
        http_line, *header_lines = data.split('\r\n')

        # parse http line
        self.cmd, self.raw_path, _httpver = http_line.split(' ', 2)
        self.httpver = float(_httpver[5:])

        # parse urlencoded args from raw_path
        if (args_offs := self.raw_path.find('?')) != -1:
            self._parse_urlencoded(self.raw_path[args_offs + 1:])
            self.path = self.raw_path[:args_offs]
        else:
            self.path = self.raw_path

        # parse header lines
        for h_key, h_val in [h.split(': ', 1) for h in header_lines]:
            self.headers[h_key] = h_val

    def _parse_multipart(self) -> None:
        """Parse multipart/form-data from the internal body."""
        boundary = self.headers['Content-Type'].split('boundary=', 1)[1]

        for param in self.body.tobytes().split(f'--{boundary}'.encode())[1:-1]:
            headers, _body = param.split(b'\r\n\r\n', 1)
            body = _body[:-2] # remove \r\n

            # find Content-Disposition
            for header in headers.decode().split('\r\n')[1:]:
                h_key, h_val = header.split(': ', 1)

                if h_key == 'Content-Disposition':
                    # find 'name' or 'filename' attribute
                    attrs = {}
                    for attr in h_val.split('; ')[1:]:
                        a_key, _a_val = attr.split('=', 1)
                        attrs[a_key] = _a_val[1:-1] # remove ""

                    if 'filename' in attrs:
                        a_val = attrs['filename']
                        self.files[a_val] = body
                        break
                    elif 'name' in attrs:
                        a_val = attrs['name']
                        self.multipart_args[a_val] = body.decode()
                        break
                    break

    async def parse(self) -> bytes:
        """Receive & parse the http request from the client."""
        loop = asyncio.get_running_loop()

        while (body_delim_offs := self._buf.find(b'\r\n\r\n')) == -1:
            self._buf += await loop.sock_recv(self.client, 1024)

        # we have all headers, parse them
        self._parse_headers(self._buf[:body_delim_offs].decode())

        if 'Content-Length' not in self.headers:
            # the request has no body to read.
            return

        content_length = int(self.headers['Content-Length'])
        to_read = ((body_delim_offs + 4) + content_length) - len(self._buf)

        if to_read:
            # there's more to read; preallocate the space
            # required and read into it from the socket.
            self._buf += b"\x00" * to_read # is this rly the fastest way?
            with memoryview(self._buf)[-to_read:] as read_view:
                while to_read:
                    nbytes = await loop.sock_recv_into(self.client, read_view)
                    read_view = read_view[nbytes:]
                    to_read -= nbytes

        # all data read from the socket, store a readonly view of the body.
        self.body = memoryview(self._buf)[body_delim_offs + 4:].toreadonly()

        if self.cmd == 'POST':
            if 'Content-Type' in self.headers:
                content_type = self.headers['Content-Type']
                if content_type.startswith('multipart/form-data'):
                    self._parse_multipart()
                elif content_type == 'application/x-www-form-urlencoded':
                    self._parse_urlencoded(self.body.obj) # hmmm

    """ Response methods """

    async def send(self, status: int, body: bytes = b'') -> None:
        """Attach appropriate headers and send data back to the client."""
        # Insert HTTP response line & content at the beginning of headers.
        header_lines = [STATUS_LINES[status]]

        if body: # Add content-length header if we are sending a body.
            header_lines.append(f'Content-Length: {len(body)}')

        # Add all user-specified response headers.
        header_lines.extend(map(': '.join, self.resp_headers.items()))

        # Create an encoded response from the headers.
        resp = ('\r\n'.join(header_lines) + '\r\n\r\n').encode()

        # Add body to response if we have one to send.
        if body:
            resp += body

        # Send all data to the client.
        loop = asyncio.get_running_loop()
        try:
            await loop.sock_sendall(self.client, resp)
        except BrokenPipeError: # TODO: detect this earlier?
            log('Connection closed by client.', Ansi.LRED)

class Route:
    """A single endpoint within of domain."""
    __slots__ = ('path', 'methods', 'handler', 'cond')
    def __init__(self, path: Union[str, Iterable, re.Pattern],
                 methods: list[str], handler: Callable) -> None:
        self.methods = methods
        self.handler = handler

        if isinstance(path, str):
            self.path = path
            self.cond = lambda k: k == path
        elif isinstance(path, Iterable):
            if isinstance(next(iter(path)), re.Pattern):
                self.cond = lambda k: any([p.match(k) for p in path])
                self.path = str(type(path)([f'~{p.pattern}' for p in path]))
            else:
                self.cond = lambda k: k in path
                self.path = str(path)
        elif isinstance(path, re.Pattern):
            self.cond = lambda k: path.match(k)
            self.path = f'~{path.pattern}' # ~ for rgx

    def __repr__(self) -> str:
        return f'{"/".join(self.methods)} {self.path}'

    def matches(self, path: str, method: str) -> bool:
        """Check if a given path & method match internals."""
        return method in self.methods and self.cond(path)

class RouteMap:
    """A collection of endpoints of a domain."""
    __slots__ = ('routes',)
    def __init__(self) -> None:
        self.routes = set() # {Route(), ...}

    def route(self, path: Hostname_Types,
              methods: list[str] = ['GET']) -> Callable:
        """Add a given route to the routemap."""
        if not isinstance(path, (str, Iterable, re.Pattern)):
            raise TypeError('Route should be str | Iterable[str] | re.Pattern')

        def wrapper(handler: Coroutine) -> Coroutine:
            self.routes.add(Route(path, methods, handler))
            return handler

        return wrapper

    def find_route(self, path: str, method: str) -> Optional[Route]:
        """Find the first route matching a given path & method."""
        for route in self.routes:
            if route.matches(path, method):
                return route

class Domain(RouteMap):
    """The main routemap, with a hostname.
       Allows for merging of additional routemaps."""
    __slots__ = ('hostname', 'cond',)
    def __init__(self, hostname: Hostname_Types) -> None:
        super().__init__()
        self.hostname = hostname # for __repr__

        if isinstance(hostname, str):
            self.cond = lambda hn: hn == hostname
        elif isinstance(hostname, Iterable):
            self.cond = lambda hn: hn in hostname
            self.hostname = str(hostname)
        elif isinstance(hostname, re.Pattern):
            self.cond = lambda hn: hostname.match(hn) is not None
            self.hostname = f'~{hostname.pattern}' # ~ for rgx
        else:
            raise TypeError('Key should be str | Iterable[str] | re.Pattern')

    def __repr__(self) -> str:
        return self.hostname

    def matches(self, hostname: str) -> bool:
        """Check if a given hostname matches our condition."""
        return self.cond(hostname)

    def add_map(self, rmap: RouteMap) -> None:
        """Add an existing routemap to our domain."""
        self.routes |= rmap.routes

class Server:
    """An asynchronous multi-domain server."""
    __slots__ = (
        'name', 'max_conns', 'gzip', 'debug',
        'sock_family', 'before_serving', 'after_serving',
        'domains', 'exceptions',
        'tasks', '_task_coros'
    )
    def __init__(self, **kwargs) -> None:
        self.name = kwargs.get('name', 'Server')
        self.max_conns = kwargs.get('max_conns', 5)
        self.gzip = kwargs.get('gzip', 0) # 0-9 valid levels
        self.debug = kwargs.get('debug', False)
        self.sock_family: Optional[socket.AddressFamily] = None

        self.before_serving: Optional[Callable] = None
        self.after_serving: Optional[Callable] = None

        self.domains = kwargs.get('domains', set())
        self.exceptions = 0 # num of exceptions handled

        self.tasks = kwargs.get('tasks', set())
        self._task_coros = kwargs.get('pending_tasks', set()) # coros not yet run

    def set_sock_mode(self, addr: Address) -> None:
        """Determine the type of socket from the address given."""
        is_inet = (
            isinstance(addr, tuple) and
            len(addr) == 2 and
            isinstance(addr[0], str) and
            isinstance(addr[1], int)
        )

        if is_inet:
            self.sock_family = socket.AF_INET
        elif isinstance(addr, str):
            self.sock_family = socket.AF_UNIX
        else:
            raise ValueError('Invalid address.')

    @property
    def using_unix_socket(self) -> bool:
        return self.sock_family is socket.AF_UNIX

    # Domain management

    def add_domain(self, domain: Domain) -> None:
        """Add a domain to the server."""
        self.domains.add(domain)

    def add_domains(self, domains: set[Domain]) -> None:
        """Add multiple domains to the server."""
        self.domains |= domains

    def remove_domain(self, domain: Domain) -> None:
        """Remove a domain from the server."""
        self.domains.remove(domain)

    def remove_domains(self, domains: set[Domain]) -> None:
        """Remove multiple domains from the server."""
        self.domains -= domains

    def find_domain(self, hostname: str):
        """Find the first domain matching a given hostname."""
        for domain in self.domains:
            if domain.matches(hostname):
                return domain

    # Task management

    def add_pending_task(self, coro: Coroutine) -> None:
        """Add a coroutine to be launched as a task at
           startup & shutdown cleanly on shutdown."""
        self._task_coros.add(coro)

    def remove_pending_task(self, coro: Coroutine) -> None:
        """Remove a pending coroutine awaiting server launch."""
        self._task_coros.remove(coro)

    def add_task(self, task: asyncio.Task) -> None:
        """Add an existing task to be cleaned
           up on shutdown."""
        self.tasks.add(task)

    def remove_task(self, task: asyncio.Task) -> None:
        """Remove an existing task from being
           cleaned up on shutdown."""
        self.tasks.remove(task)

    # True Internals

    async def dispatch(self, conn: Connection) -> int:
        """Dispatch the connection to any matching routes."""
        host = conn.headers['Host']
        path = conn.path

        resp = None

        if domain := self.find_domain(host):
            if route := domain.find_route(path, conn.cmd):
                resp = await route.handler(conn) or b''

        if resp is not None:
            if isinstance(resp, tuple):
                # code explicitly given
                code, resp = resp
            else:
                # use 200 as default
                code, resp = (200, resp)

            # gzip responses larger than a single ethernet frame
            # if it's enabled server-side and client supports it
            if (
                self.gzip > 0 and
                'Accept-Encoding' in conn.headers and
                'gzip' in conn.headers['Accept-Encoding'] and
                len(resp) > 1500 # ethernet frame size (minus headers)
            ):
                # ignore files that're already compressed heavily
                if not (
                    'Content-Type' in conn.resp_headers and
                    conn.resp_headers['Content-Type'] in (
                        # TODO: surely there's more i should be ignoring
                        'image/png', 'image/jpeg'
                    )
                ):
                    resp = gzip.compress(resp, self.gzip)
                    conn.resp_headers['Content-Encoding'] = 'gzip'
        else:
            code, resp = (404, b'Not Found.')

        await conn.send(code, resp)
        return code

    async def handle(self, client: socket.socket) -> None:
        """Handle a single client socket from the server."""
        if self.debug:
            t1 = clock_ns()

        # read & parse connection
        conn = Connection(client)
        await conn.parse()

        if self.debug:
            t2 = clock_ns()

        if 'Host' not in conn.headers:
            log('Connection missing Host header.', Ansi.LRED)
            client.shutdown(socket.SHUT_RDWR)
            client.close()
            return

        # dispatch the connection
        # to the appropriate handler
        code = await self.dispatch(conn)

        if self.debug:
            # event complete, stop timing, log result and cleanup
            t3 = clock_ns()

            t2_t1 = magnitude_fmt_time(t2 - t1)
            t3_t2 = magnitude_fmt_time(t3 - t2)

            col = (Ansi.LGREEN  if 200 <= code < 300 else
                   Ansi.LYELLOW if 300 <= code < 400 else
                   Ansi.LRED)

            uri = f'{conn.headers["Host"]}{conn.path}'

            log(f'[{conn.cmd}] {code} {uri}', col, end=' | ')
            printc(f'Parsing: {t2_t1}', Ansi.LBLUE, end=' | ')
            printc(f'Handling: {t3_t2}', Ansi.LBLUE)

        try:
            client.shutdown(socket.SHUT_RDWR)
            client.close()
        except socket.error:
            pass

    def _default_cb(self, t: asyncio.Task) -> None:
        """A simple callback for tasks to log & call exc handler."""
        if not t.cancelled():
            exc = t.exception()
            if exc and not isinstance(exc, (SystemExit, KeyboardInterrupt)):
                self.exceptions += 1

                loop = asyncio.get_running_loop()
                loop.default_exception_handler({
                    'exception': exc
                })

    def run(
        self, addr: Address,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        handle_restart: bool = True, # using USR1 signal
    ) -> None:
        """Run the server indefinitely."""
        if not loop:
            # no event loop given, check if one's running
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                # no event loop running, we need to make our own
                if spec := importlib.util.find_spec('uvloop'):
                    # use uvloop if it's already installed
                    # TODO: could make this configurable
                    # incase people want to disable it
                    # for their own use-cases?
                    uvloop = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(uvloop)

                    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

                loop = asyncio.new_event_loop()

        self.set_sock_mode(addr) # figure out family (INET4/UNIX)

        async def runner() -> None:
            log(f'=== Starting up {self.name} ===', Ansi.LMAGENTA)
            loop = asyncio.get_running_loop()

            # Call our before_serving coroutine,
            # if theres one specified.
            if self.before_serving:
                await self.before_serving()

            # Start pending coroutine tasks.
            if self.debug:
                log(f'-> Starting {len(self._task_coros)} tasks.', Ansi.LMAGENTA)

            for coro in self._task_coros:
                task = loop.create_task(coro)
                task.add_done_callback(self._default_cb) # XXX: never removed?
                self.tasks.add(task)

            self._task_coros.clear()

            # Setup socket & begin listening

            if self.using_unix_socket:
                if os.path.exists(addr):
                    os.remove(addr)

            # read/write signal listening socks
            sig_rsock, sig_wsock = os.pipe()
            os.set_blocking(sig_wsock, False)
            signal.set_wakeup_fd(sig_wsock)

            # connection listening sock
            lsock = socket.socket(self.sock_family)
            lsock.setblocking(False)

            lsock.bind(addr)
            if self.using_unix_socket:
                os.chmod(addr, 0o777)

            lsock.listen(self.max_conns)
            log(f'-> Listening @ {addr}', AnsiRGB(0x00ff7f))

            # TODO: terminal input support (tty, termios fuckery)
            # though, tbh this should be moved into gulag as it's
            # mostly a gulag-specific thing, and it'll be easier
            # to manage all the printing stuff that way.

            should_close = False
            should_restart = False

            while True:
                await asyncio.sleep(0.01) # skip loop iteration
                rlist, _, _ = select.select([lsock, sig_rsock], [], [], 0)

                for reader in rlist:
                    if reader is lsock:
                        # new connection received for server
                        client, _ = await loop.sock_accept(lsock)
                        task = loop.create_task(self.handle(client))
                        task.add_done_callback(self._default_cb)
                    elif reader is sig_rsock:
                        # received a blocked signal, shutdown
                        sig_received = signal.Signals(os.read(sig_rsock, 1)[0])
                        if sig_received is signal.SIGINT:
                            print('\x1b[2K', end='\r') # clear ^C from console
                        elif sig_received is signal.SIGUSR1:
                            should_restart = True
                        log(f'Received {signal.strsignal(sig_received)}', Ansi.LRED)
                        should_close = True
                    else:
                        raise RuntimeError(f'Unknown reader {reader}')

                if should_close:
                    break

            # server closed, clean things up.
            for sock_fd in {lsock.fileno(), sig_rsock, sig_wsock}:
                os.close(sock_fd)

            signal.set_wakeup_fd(-1)

            if self.using_unix_socket:
                os.remove(addr)

            log('-> Cancelling tasks', Ansi.LMAGENTA)
            for task in self.tasks:
                task.cancel()

            await asyncio.gather(*self.tasks, return_exceptions=True)

            if in_progress := [t for t in asyncio.all_tasks()
                               if t is not asyncio.current_task()]:
                try:
                    # allow up to 5 seconds for in-progress handlers
                    # to finish their execution, just incase they're
                    # in a half-complete state. we wouldn't want to
                    # get any sql tables into a weird state, or alike.
                    log(f'-> Awaiting {len(in_progress)} '
                        'in-progress handler(s).', Ansi.LMAGENTA)
                    await asyncio.wait(in_progress, loop=loop, timeout=5.0)
                except asyncio.TimeoutError:
                    log('-> Timed out awaiting handlers, cancelling them.', Ansi.LMAGENTA)
                    to_await = []
                    for task in in_progress:
                        if not task.cancelled():
                            task.cancel()
                            to_await.append(task)
                    await asyncio.gather(*to_await, return_exceptions=True)

            if self.after_serving:
                await self.after_serving()

            return should_restart

        # ignore any signal events in the code, the selector loop will
        # pick up on signals from the fd being awoken from set_wakeup_fd.
        def _sighandler_noop(signum, frame):
            pass

        signals = {signal.SIGINT, signal.SIGTERM, signal.SIGHUP}

        if handle_restart:
            signals.add(signal.SIGUSR1)

        for sig in signals:
            signal.signal(sig, _sighandler_noop)

        def _runner_cb(fut: asyncio.Future) -> None:
            if not fut.cancelled():
                exc = fut.exception()
                if exc and not isinstance(exc, (SystemExit, KeyboardInterrupt)):
                    loop.default_exception_handler({
                        'exception': exc
                    })

            loop.stop()

        """ run the event loop """
        future = asyncio.ensure_future(runner(), loop=loop)
        future.add_done_callback(_runner_cb)

        try:
            loop.run_forever()
        finally:
            future.remove_done_callback(_runner_cb)
            loop.close()

            log(f'=== Shut down {self.name} ===', Ansi.LMAGENTA)

            should_restart = future.result()

            if should_restart:
                log('=== Server restarting ===', Ansi.LMAGENTA)
                os.execv(sys.executable, [sys.executable] + sys.argv)



# -*- coding: utf-8 -*-

"""\
Wrappers for multiple types of osu! related files.

Source: https://github.com/cmyui/cmyui_pkg

:copyright: (c) 2021 cmyui
:license: MIT, see LICENSE for details.
"""

from .beatmap import *
from .oppai_ng import *
from .replay import *



# -*- coding: utf-8 -*-

import ctypes
import functools
from types import TracebackType
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

__all__ = ('OppaiWrapper',)

class OppaiWrapper:
    """Lightweight wrapper around franc[e]sco's c89 oppai-ng library."""
    __slots__  = ('static_lib', '_ez')

    def __init__(self, lib_path: str) -> None:
        self.static_lib = self.load_static_library(lib_path)
        self._ez = 0

    def __enter__(self) -> 'OppaiWrapper':
        self._ez = self.static_lib.ezpp_new()
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> bool:
        self.static_lib.ezpp_free(self._ez)
        self._ez = 0
        return False

    # main api

    def configure(
        self, mode: int = 0,
        acc: float = 0, mods: int = 0,
        combo: int = 0, nmiss: int = 0
    ) -> None:
        """Convenience wrapper so you don't have to
           think about the order for clobbering stuff"""
        if mode: self.set_mode(mode)
        if mods: self.set_mods(mods)
        if nmiss: self.set_nmiss(nmiss)
        if combo: self.set_combo(combo)
        if acc: self.set_accuracy_percent(acc) # n50, n100s?

    def calculate(self, osu_file_path: 'Path') -> None: # ezpp()
        osu_file_path_bytestr = str(osu_file_path).encode()
        self.static_lib.ezpp(self._ez, osu_file_path_bytestr)

    def calculate_data(self, osu_file_contents: bytes) -> None: # ezpp_data()
        self.static_lib.ezpp_data(self._ez, osu_file_contents, len(osu_file_contents))

    def calculate_dup(self, osu_file_path: 'Path') -> None: # ezpp_dup()
        osu_file_path_bytestr = str(osu_file_path).encode()
        self.static_lib.ezpp_dup(self._ez, osu_file_path_bytestr)

    def calculate_data_dup(self, osu_file_contents: bytes) -> None: # ezpp_data_dup()
        self.static_lib.ezpp_data_dup(self._ez, osu_file_contents, len(osu_file_contents))

    # get stuff

    def get_pp(self) -> float:
        return self.static_lib.ezpp_pp(self._ez)

    def get_sr(self) -> float:
        return self.static_lib.ezpp_stars(self._ez)

    def get_mode(self) -> int:
        return self.static_lib.ezpp_mode(self._ez)

    def get_combo(self) -> int:
        return self.static_lib.ezpp_combo(self._ez)

    def get_max_combo(self) -> int:
        return self.static_lib.ezpp_max_combo(self._ez)

    def get_mods(self) -> int:
        return self.static_lib.ezpp_mods(self._ez)

    def get_score_version(self) -> int:
        return self.static_lib.ezpp_score_version(self._ez)

    def get_aim_stars(self) -> float:
        return self.static_lib.ezpp_aim_stars(self._ez)

    def get_speed_stars(self) -> float:
        return self.static_lib.ezpp_speed_stars(self._ez)

    def get_aim_pp(self) -> float:
        return self.static_lib.ezpp_aim_pp(self._ez)

    def get_speed_pp(self) -> float:
        return self.static_lib.ezpp_speed_pp(self._ez)

    def get_accuracy_percent(self) -> float:
        return self.static_lib.ezpp_accuracy_percent(self._ez)

    def get_n300(self) -> int:
        return self.static_lib.ezpp_n300(self._ez)

    def get_n100(self) -> int:
        return self.static_lib.ezpp_n100(self._ez)

    def get_n50(self) -> int:
        return self.static_lib.ezpp_n50(self._ez)

    def get_nmiss(self) -> int:
        return self.static_lib.ezpp_nmiss(self._ez)

    def get_title(self) -> bytes:
        return self.static_lib.ezpp_title(self._ez)

    def get_title_unicode(self) -> bytes:
        return self.static_lib.ezpp_title_unicode(self._ez)

    def get_artist(self) -> bytes:
        return self.static_lib.ezpp_artist(self._ez)

    def get_artist_unicode(self) -> bytes:
        return self.static_lib.ezpp_artist_unicode(self._ez)

    def get_creator(self) -> bytes:
        return self.static_lib.ezpp_creator(self._ez)

    def get_version(self) -> bytes:
        return self.static_lib.ezpp_version(self._ez)

    def get_ncircles(self) -> int:
        return self.static_lib.ezpp_ncircles(self._ez)

    def get_nsliders(self) -> int:
        return self.static_lib.ezpp_nsliders(self._ez)

    def get_nspinners(self) -> int:
        return self.static_lib.ezpp_nspinners(self._ez)

    def get_nobjects(self) -> int:
        return self.static_lib.ezpp_nobjects(self._ez)

    def get_ar(self) -> float:
        return self.static_lib.ezpp_ar(self._ez)

    def get_cs(self) -> float:
        return self.static_lib.ezpp_cs(self._ez)

    def get_od(self) -> float:
        return self.static_lib.ezpp_od(self._ez)

    def get_hp(self) -> float:
        return self.static_lib.ezpp_hp(self._ez)

    def get_odms(self) -> float:
        return self.static_lib.ezpp_odms(self._ez)

    def get_autocalc(self) -> int:
        return self.static_lib.ezpp_autocalc(self._ez)

    def get_time_at(self, i: int) -> float:
        return self.static_lib.ezpp_time_at(self._ez, i)

    def get_strain_at(self, i: int, difficulty_type: int) -> float:
        return self.static_lib.ezpp_strain_at(self._ez, i, difficulty_type)

    def get_ntiming_points(self) -> int:
        return self.static_lib.ezpp_ntiming_points(self._ez)

    def get_timing_time(self, i: int) -> float:
        return self.static_lib.ezpp_timing_time(self._ez, i)

    def get_timing_ms_per_beat(self, i: int) -> float:
        return self.static_lib.ezpp_timing_ms_per_beat(self._ez, i)

    def get_timing_change(self, i: int) -> int:
        return self.static_lib.ezpp_timing_change(self._ez, i)

    # set stuff
    # NOTE: the order you call these in matters due to
    # memory clobbering (for example setting misscount
    # will reset the internal accuracy_percent). they're
    # all documented here, but you can use the configure()
    # method in the main api above if you don't want to
    # think about this level of abstraction.

    def set_aim_stars(self, aim_stars: float) -> None:
        self.static_lib.ezpp_set_aim_stars(self._ez, aim_stars)

    def set_speed_stars(self, speed_stars: float) -> None:
        self.static_lib.ezpp_set_speed_stars(self._ez, speed_stars)

    def set_base_ar(self, ar: float) -> None:
        self.static_lib.ezpp_set_base_ar(self._ez, ar)

    def set_base_od(self, od: float) -> None:
        self.static_lib.ezpp_set_base_od(self._ez, od)

    def set_base_cs(self, cs: float) -> None:
        # NOTE: will force map re-parse
        self.static_lib.ezpp_set_base_cs(self._ez, cs)

    def set_base_hp(self, hp: float) -> None:
        self.static_lib.ezpp_set_base_hp(self._ez, hp)

    def set_mode_override(self, mode: int) -> None:
        # NOTE: will force map re-parse
        self.static_lib.ezpp_set_mode_override(self._ez, mode)

    def set_mode(self, mode: int) -> None:
        self.static_lib.ezpp_set_mode(self._ez, mode)

    def set_mods(self, mods: int) -> None:
        # NOTE: will force map re-parse for
        #       hr, ez, dt, nc and ht.
        self.static_lib.ezpp_set_mods(self._ez, mods)

    def set_combo(self, combo: int) -> None:
        self.static_lib.ezpp_set_combo(self._ez, combo)

    def set_nmiss(self, nmiss: int) -> None:
        # NOTE: will force map re-parse &
        #       clobber accuracy_percent
        # (i think the map re-parse can be removed,
        #  and will talk to franc[e]sco about it)
        self.static_lib.ezpp_set_nmiss(self._ez, nmiss)

    def set_score_version(self, score_version: int) -> None:
        self.static_lib.ezpp_set_score_version(self._ez, score_version)

    def set_accuracy_percent(self, accuracy: float) -> None:
        self.static_lib.ezpp_set_accuracy_percent(self._ez, accuracy)

    def set_accuracy(self, n100: int, n50: int) -> None:
        self.static_lib.ezpp_set_accuracy(self._ez, n100, n50)

    def set_end(self, end: int) -> None:
        # NOTE: will force map re-parse &
        #       clobber accuracy_percent
        self.static_lib.ezpp_set_end(self._ez, end)

    def set_end_time(self, end_time: float) -> None:
        # NOTE: will force map re-parse &
        #       clobber accuracy_percent
        self.static_lib.ezpp_set_end_time(self._ez, end_time)

    @staticmethod
    @functools.cache
    def load_static_library(lib_path: str) -> ctypes.CDLL:
        """Load the oppai-ng static library,
           and register c types to it's api."""
        static_lib = ctypes.cdll.LoadLibrary(lib_path)

        # main api

        ezpp_new = static_lib.ezpp_new
        ezpp_new.argtypes = ()
        ezpp_new.restype = ctypes.c_int

        ezpp_free = static_lib.ezpp_free
        ezpp_free.argtypes = ()
        ezpp_free.restype = ctypes.c_void_p

        ezpp = static_lib.ezpp
        ezpp.argtypes = (ctypes.c_int, ctypes.c_char_p)
        ezpp.restype = ctypes.c_int

        ezpp_data = static_lib.ezpp_data
        ezpp_data.argtypes = (ctypes.c_int, ctypes.c_char_p, ctypes.c_int)
        ezpp_data.restype = ctypes.c_int

        ezpp_dup = static_lib.ezpp_dup
        ezpp_dup.argtypes = (ctypes.c_int, ctypes.c_char_p)
        ezpp_dup.restype = ctypes.c_int

        ezpp_data_dup = static_lib.ezpp_data_dup
        ezpp_data_dup.argtypes = (ctypes.c_int, ctypes.c_char_p, ctypes.c_int)

        # getting internals

        ezpp_pp = static_lib.ezpp_pp
        ezpp_pp.argtypes = (ctypes.c_int,)
        ezpp_pp.restype = ctypes.c_float

        ezpp_stars = static_lib.ezpp_stars
        ezpp_stars.argtypes = (ctypes.c_int,)
        ezpp_stars.restype = ctypes.c_float

        ezpp_mode = static_lib.ezpp_mode
        ezpp_mode.argtypes = (ctypes.c_int,)
        ezpp_mode.restype = ctypes.c_int

        ezpp_combo = static_lib.ezpp_combo
        ezpp_combo.argtypes = (ctypes.c_int,)
        ezpp_combo.restype = ctypes.c_int

        ezpp_max_combo = static_lib.ezpp_max_combo
        ezpp_max_combo.argtypes = (ctypes.c_int,)
        ezpp_max_combo.restype = ctypes.c_int

        ezpp_mods = static_lib.ezpp_mods
        ezpp_mods.argtypes = (ctypes.c_int,)
        ezpp_mods.restype = ctypes.c_int

        ezpp_score_version = static_lib.ezpp_score_version
        ezpp_score_version.argtypes = (ctypes.c_int,)
        ezpp_score_version.restype = ctypes.c_int

        ezpp_aim_stars = static_lib.ezpp_aim_stars
        ezpp_aim_stars.argtypes = (ctypes.c_int,)
        ezpp_aim_stars.restype = ctypes.c_float

        ezpp_speed_stars = static_lib.ezpp_speed_stars
        ezpp_speed_stars.argtypes = (ctypes.c_int,)
        ezpp_speed_stars.restype = ctypes.c_float

        ezpp_aim_pp = static_lib.ezpp_aim_pp
        ezpp_aim_pp.argtypes = (ctypes.c_int,)
        ezpp_aim_pp.restype = ctypes.c_float

        ezpp_speed_pp = static_lib.ezpp_speed_pp
        ezpp_speed_pp.argtypes = (ctypes.c_int,)
        ezpp_speed_pp.restype = ctypes.c_float

        ezpp_acc_pp = static_lib.ezpp_acc_pp
        ezpp_acc_pp.argtypes = (ctypes.c_int,)
        ezpp_acc_pp.restype = ctypes.c_float

        ezpp_accuracy_percent = static_lib.ezpp_accuracy_percent
        ezpp_accuracy_percent.argtypes = (ctypes.c_int,)
        ezpp_accuracy_percent.restype = ctypes.c_float

        ezpp_n300 = static_lib.ezpp_n300
        ezpp_n300.argtypes = (ctypes.c_int,)
        ezpp_n300.restype = ctypes.c_int

        ezpp_n100 = static_lib.ezpp_n100
        ezpp_n100.argtypes = (ctypes.c_int,)
        ezpp_n100.restype = ctypes.c_int

        ezpp_n50 = static_lib.ezpp_n50
        ezpp_n50.argtypes = (ctypes.c_int,)
        ezpp_n50.restype = ctypes.c_int

        ezpp_nmiss = static_lib.ezpp_nmiss
        ezpp_nmiss.argtypes = (ctypes.c_int,)
        ezpp_nmiss.restype = ctypes.c_int

        ezpp_title = static_lib.ezpp_title
        ezpp_title.argtypes = (ctypes.c_int,)
        ezpp_title.restype = ctypes.c_char_p

        ezpp_title_unicode = static_lib.ezpp_title_unicode
        ezpp_title_unicode.argtypes = (ctypes.c_int,)
        ezpp_title_unicode.restype = ctypes.c_char_p

        ezpp_artist = static_lib.ezpp_artist
        ezpp_artist.argtypes = (ctypes.c_int,)
        ezpp_artist.restype = ctypes.c_char_p

        ezpp_artist_unicode = static_lib.ezpp_artist_unicode
        ezpp_artist_unicode.argtypes = (ctypes.c_int,)
        ezpp_artist_unicode.restype = ctypes.c_char_p

        ezpp_creator = static_lib.ezpp_creator
        ezpp_creator.argtypes = (ctypes.c_int,)
        ezpp_creator.restype = ctypes.c_char_p

        ezpp_version = static_lib.ezpp_version
        ezpp_version.argtypes = (ctypes.c_int,)
        ezpp_version.restype = ctypes.c_char_p

        ezpp_ncircles = static_lib.ezpp_ncircles
        ezpp_ncircles.argtypes = (ctypes.c_int,)
        ezpp_ncircles.restype = ctypes.c_int

        ezpp_nsliders = static_lib.ezpp_nsliders
        ezpp_nsliders.argtypes = (ctypes.c_int,)
        ezpp_nsliders.restype = ctypes.c_int

        ezpp_nspinners = static_lib.ezpp_nspinners
        ezpp_nspinners.argtypes = (ctypes.c_int,)
        ezpp_nspinners.restype = ctypes.c_int

        ezpp_nobjects = static_lib.ezpp_nobjects
        ezpp_nobjects.argtypes = (ctypes.c_int,)
        ezpp_nobjects.restype = ctypes.c_int

        ezpp_ar = static_lib.ezpp_ar
        ezpp_ar.argtypes = (ctypes.c_int,)
        ezpp_ar.restype = ctypes.c_float

        ezpp_cs = static_lib.ezpp_cs
        ezpp_cs.argtypes = (ctypes.c_int,)
        ezpp_cs.restype = ctypes.c_float

        ezpp_od = static_lib.ezpp_od
        ezpp_od.argtypes = (ctypes.c_int,)
        ezpp_od.restype = ctypes.c_float

        ezpp_hp = static_lib.ezpp_hp
        ezpp_hp.argtypes = (ctypes.c_int,)
        ezpp_hp.restype = ctypes.c_float

        ezpp_odms = static_lib.ezpp_odms
        ezpp_odms.argtypes = (ctypes.c_int,)
        ezpp_odms.restype = ctypes.c_float

        ezpp_autocalc = static_lib.ezpp_autocalc
        ezpp_autocalc.argtypes = (ctypes.c_int,)
        ezpp_autocalc.restype = ctypes.c_int

        ezpp_time_at = static_lib.ezpp_time_at
        ezpp_time_at.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_time_at.restype = ctypes.c_float

        ezpp_strain_at = static_lib.ezpp_strain_at
        ezpp_strain_at.argtypes = (ctypes.c_int, ctypes.c_int, ctypes.c_int)
        ezpp_strain_at.restype = ctypes.c_float

        ezpp_strain_at = static_lib.ezpp_strain_at
        ezpp_strain_at.argtypes = (ctypes.c_int, ctypes.c_int, ctypes.c_int)
        ezpp_strain_at.restype = ctypes.c_float

        ezpp_ntiming_points = static_lib.ezpp_ntiming_points
        ezpp_ntiming_points.argtypes = (ctypes.c_int,)
        ezpp_ntiming_points.restype = ctypes.c_int

        ezpp_timing_time = static_lib.ezpp_timing_time
        ezpp_timing_time.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_timing_time.restype = ctypes.c_float

        ezpp_timing_ms_per_beat = static_lib.ezpp_timing_ms_per_beat
        ezpp_timing_ms_per_beat.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_timing_ms_per_beat.restype = ctypes.c_float

        ezpp_timing_change = static_lib.ezpp_timing_change
        ezpp_timing_change.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_timing_change.restype = ctypes.c_int

        # setting internals

        ezpp_set_aim_stars = static_lib.ezpp_set_aim_stars
        ezpp_set_aim_stars.argtypes = (ctypes.c_int, ctypes.c_float)
        ezpp_set_aim_stars.restype = ctypes.c_void_p

        ezpp_set_speed_stars = static_lib.ezpp_set_speed_stars
        ezpp_set_speed_stars.argtypes = (ctypes.c_int, ctypes.c_float)
        ezpp_set_speed_stars.restype = ctypes.c_void_p

        ezpp_set_base_ar = static_lib.ezpp_set_base_ar
        ezpp_set_base_ar.argtypes = (ctypes.c_int, ctypes.c_float)
        ezpp_set_base_ar.restype = ctypes.c_void_p

        ezpp_set_base_od = static_lib.ezpp_set_base_od
        ezpp_set_base_od.argtypes = (ctypes.c_int, ctypes.c_float)
        ezpp_set_base_od.restype = ctypes.c_void_p

        ezpp_set_base_hp = static_lib.ezpp_set_base_hp
        ezpp_set_base_hp.argtypes = (ctypes.c_int, ctypes.c_float)
        ezpp_set_base_hp.restype = ctypes.c_void_p

        ezpp_set_mode = static_lib.ezpp_set_mode
        ezpp_set_mode.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_set_mode.restype = ctypes.c_void_p

        ezpp_set_combo = static_lib.ezpp_set_combo
        ezpp_set_combo.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_set_combo.restype = ctypes.c_void_p

        ezpp_set_score_version = static_lib.ezpp_set_score_version
        ezpp_set_score_version.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_set_score_version.restype = ctypes.c_void_p

        ezpp_set_accuracy_percent = static_lib.ezpp_set_accuracy_percent
        ezpp_set_accuracy_percent.argtypes = (ctypes.c_int, ctypes.c_float)
        ezpp_set_accuracy_percent.restype = ctypes.c_void_p

        ezpp_set_autocalc = static_lib.ezpp_set_autocalc
        ezpp_set_autocalc.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_set_autocalc.restype = ctypes.c_void_p

        # forces map re-parse for map-changing mods
        # (this is an implementation detail of oppai-ng)

        ezpp_set_mods = static_lib.ezpp_set_mods
        ezpp_set_mods.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_set_mods.restype = ctypes.c_void_p

        # forces map re-parse

        ezpp_set_base_cs = static_lib.ezpp_set_base_cs
        ezpp_set_base_cs.argtypes = (ctypes.c_int, ctypes.c_float)
        ezpp_set_base_cs.restype = ctypes.c_void_p

        ezpp_set_mode_override = static_lib.ezpp_set_mode_override
        ezpp_set_mode_override.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_set_mode_override.restype = ctypes.c_void_p

        # forces map re-parse & clobbers acc

        ezpp_set_nmiss = static_lib.ezpp_set_nmiss
        ezpp_set_nmiss.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_set_nmiss.restype = ctypes.c_void_p

        ezpp_set_end = static_lib.ezpp_set_end
        ezpp_set_end.argtypes = (ctypes.c_int, ctypes.c_int)
        ezpp_set_end.restype = ctypes.c_void_p

        ezpp_set_end_time = static_lib.ezpp_set_end_time
        ezpp_set_end_time.argtypes = (ctypes.c_int, ctypes.c_float)
        ezpp_set_end_time.restype = ctypes.c_void_p

        ezpp_set_accuracy = static_lib.ezpp_set_accuracy
        ezpp_set_accuracy.argtypes = (ctypes.c_int, ctypes.c_int, ctypes.c_int)
        ezpp_set_accuracy.restype = ctypes.c_void_p

        return static_lib



# -*- coding: utf-8 -*-

import colorsys
import random
import string
from collections import namedtuple
from datetime import datetime
from datetime import tzinfo
from enum import IntEnum
from functools import wraps
from time import time_ns
from typing import Callable
from typing import Optional
from typing import Union

__all__ = ('get_timestamp', '_isdecimal', 'rstring',
           'async_cache', 'TimeScale', 'timef',
           'rainbow_color_stops',
           'magnitude_fmt_time')

ts_fmt = ('%I:%M:%S%p', '%d/%m/%Y %I:%M:%S%p')
def get_timestamp(full: bool = False, tz: Optional[tzinfo] = None) -> str:
    return f'{datetime.now(tz=tz):{ts_fmt[full]}}'

def _isdecimal(s: str, _float: bool = False,
               _negative: bool = False) -> None:
    if _float:
        s = s.replace('.', '', 1)

    if _negative:
        s = s.replace('-', '', 1)

    return s.isdecimal()

__chars = string.ascii_letters + string.digits
def rstring(l: int, seq: str = __chars) -> str:
    return ''.join((random.choice(seq) for _ in range(l)))

_CacheStats = namedtuple('CacheStats', ['hits', 'misses'])

def _make_key(args, kwargs):
    """Get a hashed value out of args & kwargs."""
    key = args

    if kwargs:
        for item in kwargs.items():
            key += item

    return hash(key)

def async_cache(f):
    """Lightweight unbounded async function cache."""
    cache = {}
    hits = misses = 0

    @wraps(f)
    async def wrapper(*args, **kwargs):
        key = _make_key(args, kwargs)

        if key in cache:
            nonlocal hits
            hits += 1

            val = cache[key]
        else:
            nonlocal misses
            misses += 1

            val = await f(*args, **kwargs)
            cache[key] = val

        return val

    def cache_clear():
        nonlocal hits, misses
        hits = misses = 0
        cache.clear()

    def cache_stats():
        nonlocal hits, misses
        return _CacheStats(hits, misses)

    wrapper.cache_clear = cache_clear
    wrapper.cache_stats = cache_stats
    return wrapper

class TimeScale(IntEnum):
    Years = 0
    Months = 1
    Days = 2
    Hours = 3
    Minutes = 4
    Seconds = 5
    Milliseconds = 6
    Microseconds = 7
    Nanoseconds = 8

Call = namedtuple('CallEvent', ['duration', 'time',
                                'args', 'kwargs'])

timescale_divisor = lambda ts: { # lol
    TimeScale.Years: 1000 * 1000 * 1000 * 60 * 60 * 24 * 365,
    TimeScale.Months: 1000 * 1000 * 1000 * 60 * 60 * 24 * 30,
    TimeScale.Days: 1000 * 1000 * 1000 * 60 * 60 * 24,
    TimeScale.Hours: 1000 * 1000 * 1000 * 60 * 60,
    TimeScale.Minutes: 1000 * 1000 * 1000 * 60,
    TimeScale.Seconds: 1000 * 1000 * 1000,
    TimeScale.Milliseconds: 1000 * 1000,
    TimeScale.Microseconds: 1000,
    TimeScale.Nanoseconds: 1
}[ts]

def timef(f: Callable):
    calls: list[Call] = []

    @wraps(f)
    def wrapper(*args, **kwargs):
        # all timing is done in nanoseconds,
        # conversion will be done later when
        # displaying data.
        start = time_ns() # start clock
        result = f(*args, **kwargs) # run func
        end = time_ns() # stop clock

        duration = end - start

        # add to times log
        nonlocal calls
        calls.append(Call(duration, end, args, kwargs))

        return result

    def average(ts: TimeScale = TimeScale.Milliseconds):
        nonlocal calls
        total = sum([c.duration for c in calls])
        average = total / len(calls)
        return average / timescale_divisor(ts)

    wrapper.calls = calls # maybe temp?
    wrapper.average = average
    return wrapper

def rainbow_color_stops(
    n: int = 10,
    lum: float = 0.5,
    end: float = 2 / 3
) -> list[tuple[int, int, int]]:
    # https://stackoverflow.com/a/58811633
    return [(r * 255, g * 255, b * 255)
            for r, g, b in [colorsys.hls_to_rgb(end * i / (n - 1), lum, 1)
                            for i in range(n)]]

# TODO: genericize this to metric all units?
TIME_ORDER_SUFFIXES = ['nsec', 'μsec', 'msec', 'sec']
def magnitude_fmt_time(
    t: Union[int, float] # in nanosec
) -> str:
    for suffix in TIME_ORDER_SUFFIXES:
        if t < 1000:
            break
        t /= 1000
    return f'{t:.2f} {suffix}'



# -*- coding: utf-8 -*-

from typing import Any
from typing import AsyncGenerator
from typing import Optional
from typing import Sequence
from typing import Union

import aiomysql
import mysql.connector.pooling

__all__ = (
    # Informational
    'SQLParams',
    'DictSQLResult',
    'TupleSQLResult',

    # Functional
    'AsyncSQLPool',
    'SQLPool'
)

SQLParams = Sequence[Any]
DictSQLResult = Optional[dict[str, Any]]
TupleSQLResult = Optional[tuple[Any, ...]]

class AsyncSQLPool:
    """A thin wrapper around an asynchronous mysql pool
       for single query connections."""
    __slots__ = ('pool',)

    def __init__(self):
        self.pool: Optional[aiomysql.Pool] = None

    async def connect(self, config: dict[str, object]) -> None:
        """Connect to the mysql server with a given config."""
        self.pool = await aiomysql.create_pool(**config, autocommit=True)

    async def close(self) -> None:
        """Close active connection to the mysql server."""
        self.pool.close()
        await self.pool.wait_closed()

    async def execute(self, query: str, params: SQLParams = []) -> int:
        """Acquire a connection & execute a given querystring."""
        async with self.pool.acquire() as conn:
            async with conn.cursor(aiomysql.Cursor) as cur:
                await cur.execute(query, params)
                await conn.commit()

                return cur.lastrowid

    async def fetch(
        self, query: str,
        params: SQLParams = [],
        _all: bool = False,
        _dict: bool = True
    ) -> Union[DictSQLResult, TupleSQLResult]:
        """Acquire a connection & fetch first result
           row for a given querystring."""
        cursor_type = (aiomysql.DictCursor if _dict else
                       aiomysql.Cursor)

        async with self.pool.acquire() as conn:
            async with conn.cursor(cursor_type) as cur:
                await cur.execute(query, params)
                return await (cur.fetchall if _all else cur.fetchone)()

    async def fetchall(
        self, query: str,
        params: SQLParams = [],
        _dict: bool = True
    ) -> tuple[Union[DictSQLResult, TupleSQLResult], ...]:
        """Acquire a connection & fetch all result
           rows for a given querystring."""
        return await self.fetch(query, params, _all=True, _dict=_dict)

    async def iterall(
        self, query: str,
        params: SQLParams = [],
        _dict: bool = True
    ) -> AsyncGenerator[Union[DictSQLResult, TupleSQLResult], None]:
        """Like fetchall, but returns an async generator."""
        cursor_type = (aiomysql.DictCursor if _dict else
                       aiomysql.Cursor)

        async with self.pool.acquire() as conn:
            async with conn.cursor(cursor_type) as cur:
                await cur.execute(query, params)

                async for row in cur:
                    yield row

# NOTE: i don't really use this anymore
class SQLPool(mysql.connector.pooling.MySQLConnectionPool):
    """A thin wrapper around a mysql pool for single query connections."""
    def execute(self, query: str, params: SQLParams = []) -> int:
        """Acquire a connection & execute a given querystring."""
        if not (cnx := self.get_connection()):
            raise Exception('mysql: failed to retrieve a worker for cmyui.execute()')

        cur = cnx.cursor()
        cur.execute(query, params)
        cur.fetchmany()

        # Since we are executing a command, we
        # simply return the last row affected's id.
        res = cur.lastrowid

        cur.close()
        cnx.close()
        return res

    def fetch(
        self, query: str,
        params: SQLParams = [],
       _all: bool = False,
       _dict: bool = True
    ) -> Union[DictSQLResult, TupleSQLResult]:
        """Acquire a connection & fetch first result
           row for a given querystring."""
        if not (cnx := self.conn.get_connection()):
            raise Exception('mysql: failed to retrieve a worker for cmyui.fetch()')

        cur = cnx.cursor(dictionary=_dict, buffered=True)
        cur.execute(query, params)

        # We are fetching data.
        res = (cur.fetchall if _all else cur.fetchone)()

        cur.close()
        cnx.close()
        return res

    def fetchall(
        self, query: str,
        params: SQLParams = [],
        _dict: bool = True
    ) -> tuple[Union[DictSQLResult, TupleSQLResult], ...]:
        """Acquire a connection & fetch all result rows for a given querystring."""
        return self.fetch(query, params, _all=True, _dict=_dict)



# -*- coding: utf-8 -*-

import functools

__all__ = ('Version',)

@functools.total_ordering
class Version:
    __slots__ = ('major', 'minor', 'micro')
    def __init__(self, major: int, minor: int, micro: int) -> None:
        self.major = major
        self.minor = minor
        self.micro = micro

    def __repr__(self) -> str:
        return f'{self.major}.{self.minor}.{self.micro}'

    def __hash__(self) -> int:
        return self.as_tuple.__hash__()

    def __eq__(self, other: 'Version') -> bool:
        return self.as_tuple == other.as_tuple

    def __lt__(self, other: 'Version') -> bool:
        return self.as_tuple < other.as_tuple

    @property
    def as_tuple(self) -> tuple[int]:
        return (self.major, self.minor, self.micro)

    @classmethod
    def from_str(cls, s: str) -> 'Version':
        if len(split := s.split('.')) != 3:
            return

        return cls(*map(int, split))
